import multiprocessing as mp
import os
import signal
import time
from enum import Enum
from multiprocessing import Manager
from typing import Callable

import newrelic.agent
import sentry_sdk
from sentry_sdk import capture_exception

from mage_ai.orchestration.db.process import start_session_and_run
from mage_ai.orchestration.queue.config import QueueConfig
from mage_ai.orchestration.queue.queue import Queue
from mage_ai.services.newrelic import initialize_new_relic
from mage_ai.services.redis.redis import init_redis_client
from mage_ai.settings import HOSTNAME, REDIS_URL, SENTRY_DSN, SENTRY_TRACES_SAMPLE_RATE

LIVENESS_TIMEOUT_SECONDS = 300


class JobStatus(str, Enum):
    QUEUED = 'queued'
    RUNNING = 'running'  # Not used. The value for RUNNING job is process id.
    COMPLETED = 'completed'
    CANCELLED = 'cancelled'


class ProcessQueue(Queue):
    def __init__(self, queue_config: QueueConfig):
        """
        A process-based implementation of a queue that allows enqueueing and executing jobs using
        multiprocessing.

        Args:
            queue_config (QueueConfig): The configuration for the process queue.

        Attributes:
            queue_config (QueueConfig): The configuration for the process queue.
            queue (mp.Queue): A multiprocessing queue for storing the jobs.
            size (int): The size of the worker pool (defaults to the number of CPUs).
            mp_manager (Manager): A multiprocessing manager for maintaining a shared dictionary for
                jobs.

        """
        self.queue_config = queue_config
        self.process_queue_config = self.queue_config.process_queue_config
        self.queue = mp.Queue()
        self.size = queue_config.concurrency or os.cpu_count()
        self.mp_manager = Manager()
        self.job_dict = self.mp_manager.dict()

        # Initialize redis client to track jobs across multiple replicas
        if self.process_queue_config and self.process_queue_config.redis_url:
            redis_url = self.process_queue_config.redis_url
        elif REDIS_URL:
            redis_url = REDIS_URL
        else:
            redis_url = None

        self.redis_client = init_redis_client(redis_url)

        self.client_id = f'HOST_{HOSTNAME}_PID_{os.getpid()}'

        self.worker_pool_proc = None

    def clean_up_jobs(self):
        """
        Cleans up completed jobs from the job dictionary.
        """
        job_ids = self.job_dict.keys()
        for job_id in job_ids:
            if job_id in self.job_dict and not self.has_job(job_id):
                del self.job_dict[job_id]

    def enqueue(self, job_id: str, target: Callable, *args, **kwargs):
        """
        Enqueues a job to be executed in the worker pool.

        Args:
            job_id (str): The ID of the job.
            target (Callable): The target function to execute.
            *args: Variable length argument list for the target function.
            **kwargs: Keyword arguments for the target function.

        """
        if self.has_job(job_id):
            self._print(f'Job {job_id} exists. Skip enqueue.')
            return
        self._print(f'Enqueue job {job_id}')
        if self.redis_client:
            self.redis_client.set(job_id, self.client_id)
        if self.redis_client:
            self.redis_client.set(self.client_id, '1', ex=LIVENESS_TIMEOUT_SECONDS)
        self.queue.put([job_id, target, args, kwargs])
        self.job_dict[job_id] = JobStatus.QUEUED
        if not self.is_worker_pool_alive():
            self.start_worker_pool()

    def has_job(self, job_id: str) -> bool:
        """
        Checks if a job with the given ID exists in the queue or is currently being executed.

        Args:
            job_id (str): The ID of the job.

        Returns:
            bool: True if the job exists, False otherwise.

        """
        if self.redis_client:
            job_client_id = self.redis_client.get(job_id)
            if not job_client_id:
                return False
            if job_client_id != self.client_id and self.redis_client.get(job_client_id):
                return True
        job = self.job_dict.get(job_id)
        return job is not None and (job == JobStatus.QUEUED or isinstance(job, int))

    def kill_job(self, job_id: str):
        """
        Cancels and kills a job with the given ID if it is running.

        Args:
            job_id (str): The ID of the job.

        """
        print(f'Kill job {job_id}, job_dict {self.job_dict}')
        job = self.job_dict.get(job_id)
        if not job:
            return
        if isinstance(job, int):
            if job == os.getpid():
                # Update the job status before the process is killed
                self.job_dict[job_id] = JobStatus.CANCELLED
            try:
                os.kill(job, signal.SIGKILL)
            except Exception as err:
                print(err)
        self.job_dict[job_id] = JobStatus.CANCELLED

    def start_worker_pool(self):
        """
        Starts the worker pool by creating a new process for executing jobs.
        """
        self.worker_pool_proc = mp.Process(
            target=poll_job_and_execute,
            args=[
                self.queue,
                self.size,
                self.job_dict,
                self.redis_client,
                self.client_id,
            ],
        )
        self.worker_pool_proc.start()

    def is_worker_pool_alive(self) -> bool:
        """
        Checks if the worker pool process is alive.

        Returns:
            bool: True if the worker pool process is alive, False otherwise.

        """
        if self.worker_pool_proc is None:
            return False
        return self.worker_pool_proc.is_alive()


class Worker(mp.Process):
    def __init__(
        self,
        queue: mp.Queue,
        job_dict,
    ):
        """
        A worker process for executing jobs from the process queue.

        Args:
            queue (mp.Queue): The multiprocessing queue from which jobs are fetched.
            job_dict: The shared job dictionary.

        Attributes:
            queue (mp.Queue): The multiprocessing queue from which jobs are fetched.
            job_dict: The shared job dictionary.
            dsn (str): The Sentry DSN for error reporting.

        """
        super().__init__()
        self.queue = queue
        self.job_dict = job_dict
        self.dsn = SENTRY_DSN
        if self.dsn:
            sentry_sdk.init(
                self.dsn,
                traces_sample_rate=SENTRY_TRACES_SAMPLE_RATE,
            )
        initialize_new_relic()

    @newrelic.agent.background_task(name='worker-run', group='Task')
    def run(self):
        """
        The entry point for the worker process.

        Fetches a job from the queue, executes it, and updates the job status in the job dictionary.

        """
        if not self.queue.empty():
            args = self.queue.get()
            job_id = args[0]
            print(f'Run worker for job {job_id}')
            if self.job_dict[job_id] != JobStatus.QUEUED:
                return
            self.job_dict[job_id] = self.pid
            try:
                start_session_and_run(args[1], *args[2], **args[3])
            except Exception as e:
                if self.dsn:
                    capture_exception(e)
                raise
            finally:
                self.job_dict[job_id] = JobStatus.COMPLETED


def poll_job_and_execute(
    queue: mp.Queue,
    size: int,
    job_dict,
    redis_client,
    client_id: str,
):
    """
    Continuously polls the job queue and executes jobs in a worker pool.

    Args:
        queue: The multiprocessing queue from which jobs are fetched.
        size: The size of the worker pool.
        job_dict: The shared job dictionary.

    """
    workers = []
    while True:
        workers = [w for w in workers if w.is_alive()]
        print(f'Worker pool size: {len(workers)}')
        if not workers and queue.empty():
            break
        while not queue.empty():
            if len(workers) >= size:
                break
            worker = Worker(queue, job_dict)
            worker.start()
            workers.append(worker)
        time.sleep(1)
        if redis_client and client_id:
            redis_client.set(client_id, '1', ex=LIVENESS_TIMEOUT_SECONDS)



import newrelic.agent

from mage_ai.settings import ENABLE_NEW_RELIC, NEW_RELIC_CONFIG_PATH


def initialize_new_relic():
    enable_new_relic = ENABLE_NEW_RELIC
    application = None
    if ENABLE_NEW_RELIC:
        try:
            newrelic.agent.initialize(NEW_RELIC_CONFIG_PATH)
            application = newrelic.agent.register_application(timeout=10)
        except newrelic.api.exceptions.ConfigurationError as error:
            print(f"Configuration error with new relic initialization. Disable "
                  f"new relic reporting. Message: {error}")
            enable_new_relic = False
        except Exception as error:
            print("Unexpected error with new relic initialization. Disable "
                  f"new relic reporting. Message: {error}")
            enable_new_relic = False
    return (enable_new_relic, application)



from typing import Any, Dict, List

from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.block.sql.utils.shared import (
    create_upstream_block_tables as create_upstream_block_tables_orig,
)
from mage_ai.data_preparation.models.block.sql.utils.shared import (
    has_create_or_insert_statement,
    interpolate_input,
    split_query_string,
)
from mage_ai.io.config import ConfigKey


def create_upstream_block_tables(
    loader,
    block,
    cascade_on_drop: bool = False,
    configuration: Dict = None,
    execution_partition: str = None,
    cache_upstream_dbt_models: bool = False,
    query: str = None,
    dynamic_block_index: int = None,
    dynamic_upstream_block_uuids: List[str] = None,
):
    create_upstream_block_tables_orig(
        loader,
        block,
        cascade_on_drop,
        configuration,
        execution_partition,
        cache_upstream_dbt_models,
        cache_keys=[
            ConfigKey.MSSQL_DATABASE,
            ConfigKey.MSSQL_SCHEMA,
            ConfigKey.MSSQL_HOST,
            ConfigKey.MSSQL_PORT,
        ],
        no_schema=False,
        query=query,
        schema_name=loader.default_schema(),
        dynamic_block_index=dynamic_block_index,
        dynamic_upstream_block_uuids=dynamic_upstream_block_uuids,
    )


def interpolate_input_data(
    block,
    query: str,
    dynamic_block_index: int = None,
    dynamic_upstream_block_uuids: List[str] = None,
):
    return interpolate_input(
        block,
        query,
        lambda db, schema, tn: tn,
        dynamic_block_index=dynamic_block_index,
        dynamic_upstream_block_uuids=dynamic_upstream_block_uuids,
    )


def execute_raw_sql(
    loader,
    block: Block,
    query_string: str,
    configuration: Dict = None,
    should_query: bool = False,
) -> List[Any]:
    if configuration is None:
        configuration = {}
    queries = []
    fetch_query_at_indexes = []

    has_create_or_insert = has_create_or_insert_statement(query_string)

    for query in split_query_string(query_string):
        if has_create_or_insert:
            queries.append(query)
            fetch_query_at_indexes.append(False)
        else:
            queries.append(query)
            fetch_query_at_indexes.append(True)

    if should_query and has_create_or_insert:
        queries.append(f'SELECT TOP(1000) * FROM {block.full_table_name}')
        fetch_query_at_indexes.append(block.full_table_name)

    results = loader.execute_queries(
        queries,
        commit=True,
        fetch_query_at_indexes=fetch_query_at_indexes,
    )

    if should_query:
        return [results[-1]]

    return []



from datetime import datetime, timedelta
from typing import Dict, List

import dateutil.parser
from dateutil.relativedelta import relativedelta

from mage_ai.data_preparation.models.triggers import ScheduleInterval, ScheduleStatus
from mage_ai.orchestration.db.models.schedules import (
    Backfill,
    PipelineRun,
    PipelineSchedule,
)
from mage_ai.shared.hash import merge_dict


def start_backfill(backfill: Backfill) -> List[PipelineRun]:
    pipeline_runs = []
    variables_list = __build_variables_list(backfill)

    if len(variables_list) == 0:
        return []

    backfill_variables = backfill.variables

    pipeline_schedule = backfill.pipeline_schedule
    if not pipeline_schedule:
        pipeline_schedule = PipelineSchedule.create(
            name=f'Backfill {backfill.name}',
            pipeline_uuid=backfill.pipeline_uuid,
            schedule_interval=ScheduleInterval.ONCE,
            start_time=datetime.utcnow(),
            variables=backfill_variables,
        )

    for backfill_run_variables in variables_list:
        execution_date = None
        if 'execution_date' in backfill_run_variables:
            execution_date = dateutil.parser.parse(backfill_run_variables['execution_date'])

        pipeline_run = PipelineRun.create(
            backfill_id=backfill.id,
            execution_date=execution_date,
            pipeline_schedule_id=pipeline_schedule.id,
            pipeline_uuid=backfill.pipeline_uuid,
            variables=merge_dict(backfill_variables or {}, backfill_run_variables),
        )
        pipeline_runs.append(pipeline_run)

    backfill.update(pipeline_schedule_id=pipeline_schedule.id)
    pipeline_schedule.update(status=ScheduleStatus.ACTIVE)

    return pipeline_runs


def preview_run_dates(backfill: Backfill) -> List[Dict]:
    variables_list = __build_variables_list(backfill)

    return variables_list


def cancel_backfill(backfill: Backfill) -> None:
    for pipeline_run in PipelineRun.query.filter(
        PipelineRun.backfill_id == backfill.id,
        PipelineRun.status.not_in([
            PipelineRun.PipelineRunStatus.CANCELLED,
            PipelineRun.PipelineRunStatus.COMPLETED,
            PipelineRun.PipelineRunStatus.FAILED,
        ]),
    ):
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.CANCELLED)

    backfill.update(status=Backfill.Status.CANCELLED)
    backfill.pipeline_schedule.update(status=ScheduleStatus.INACTIVE)


def __build_variables_list(backfill: Backfill) -> List[Dict]:
    if backfill.block_uuid:
        return []

    dates = __build_dates(backfill)
    number_of_dates = len(dates)

    arr = []

    for idx, execution_date in enumerate(dates):
        interval_end_datetime = None
        interval_seconds = None
        interval_start_datetime = execution_date
        interval_start_datetime_previous = None

        if idx < number_of_dates - 1:
            interval_end_datetime = dates[idx + 1]
            interval_seconds = (
                interval_end_datetime.timestamp() - interval_start_datetime.timestamp()
            )
        elif idx >= 1 and idx == number_of_dates - 1:
            # Last date
            interval_start_datetime_previous = dates[idx - 1]
            interval_seconds = (
                interval_start_datetime.timestamp() - interval_start_datetime_previous.timestamp()
            )
            interval_end_datetime = interval_start_datetime + timedelta(seconds=interval_seconds)

        if not interval_start_datetime_previous and (interval_seconds and interval_start_datetime):
            interval_start_datetime_previous = interval_start_datetime - timedelta(
                seconds=interval_seconds,
            )

        if interval_end_datetime:
            interval_end_datetime = interval_end_datetime.isoformat()

        if interval_start_datetime:
            interval_start_datetime = interval_start_datetime.isoformat()

        if interval_start_datetime_previous:
            interval_start_datetime_previous = interval_start_datetime_previous.isoformat()

        arr.append(dict(
            ds=execution_date.strftime('%Y-%m-%d'),
            execution_date=execution_date.isoformat(),
            hr=execution_date.strftime('%H'),
            interval_end_datetime=interval_end_datetime,
            interval_seconds=interval_seconds,
            interval_start_datetime=interval_start_datetime,
            interval_start_datetime_previous=interval_start_datetime_previous,
        ))

    return arr


def __build_dates(backfill: Backfill) -> List[datetime]:
    dates = []

    current_datetime = backfill.start_datetime
    while current_datetime <= backfill.end_datetime:
        dates.append(current_datetime)

        interval_type = backfill.interval_type
        if interval_type not in [
            Backfill.IntervalType.MONTH,
            Backfill.IntervalType.YEAR,
        ]:
            unit = f'{interval_type.value}s'
            if Backfill.IntervalType.CUSTOM == interval_type:
                unit = 'seconds'
            current_datetime += timedelta(**{unit: backfill.interval_units})
        else:
            key = None
            if Backfill.IntervalType.MONTH == interval_type:
                key = 'months'
            elif Backfill.IntervalType.YEAR == interval_type:
                key = 'years'

            delta = {
                key: backfill.interval_units,
            }
            current_datetime = current_datetime + relativedelta(**delta)
            current_datetime = current_datetime.replace(tzinfo=current_datetime.tzinfo)

    return dates



import asyncio
import os
from datetime import datetime
from typing import Dict, Union

from mage_ai.cache.base import BaseCache
from mage_ai.cache.constants import CACHE_KEY_BLOCKS_TO_PIPELINE_MAPPING
from mage_ai.cache.utils import build_pipeline_dict


class BlockCache(BaseCache):
    cache_key = CACHE_KEY_BLOCKS_TO_PIPELINE_MAPPING

    @classmethod
    async def initialize_cache(self, replace: bool = False) -> 'BlockCache':
        cache = self()
        if replace or not cache.exists():
            await cache.initialize_cache_for_all_pipelines()

        return cache

    def build_key(self, block: Union[Dict]) -> str:
        """Generate cache key for block.

        Args:
            block (Union[Dict]): The block dict or object.

        Returns:
            str: The cache key generated with block_type and block_uuid.
                 If block_type or block_uuid is None, return None.
        """
        block_type = ''
        block_uuid = ''

        if type(block) is dict:
            block_type = block.get('type')
            block_uuid = block.get('uuid')
        else:
            block_type = block.type
            block_uuid = block.uuid
        if not block_type or not block_uuid:
            return None
        return os.path.join(block_type, block_uuid)

    def get_pipelines(self, block) -> Dict:
        pipelines_dict = {}

        mapping = self.get(self.cache_key)
        if mapping is not None:
            key = self.build_key(block)
            if key:
                pipelines_dict = mapping.get(key, {})

        return pipelines_dict

    def add_pipeline(self, block, pipeline) -> None:
        self.update_pipeline(block, pipeline, added_at=datetime.utcnow().timestamp())

    def move_pipelines(self, new_block, old_block) -> None:
        new_key = self.build_key(new_block)
        if not new_key:
            return

        old_key = self.build_key(old_block)
        if not old_key:
            return

        mapping = self.get(self.cache_key)
        if mapping is None:
            mapping = {}

        pipelines_dict = mapping.get(old_key, {})
        mapping[new_key] = pipelines_dict
        mapping.pop(old_key, None)

        self.set(self.cache_key, mapping)

    def update_pipeline(
        self,
        block,
        pipeline,
        added_at: str = None,
    ) -> None:
        self.update_pipelines(block, [pipeline], added_at)

    def update_pipelines(
        self,
        block,
        pipelines,
        added_at: str = None,
    ) -> None:
        key = self.build_key(block)
        if not key:
            return

        mapping = self.get(self.cache_key)
        if mapping is None:
            mapping = {}

        pipelines_dict = mapping.get(key, {})

        for pipeline in pipelines:
            pipeline_uuid = pipeline.get('uuid') if type(pipeline) is dict else pipeline.uuid
            pipelines_dict[pipeline_uuid] = build_pipeline_dict(
                pipeline,
                added_at=added_at,
            )

        mapping[key] = pipelines_dict

        self.set(self.cache_key, mapping)

    def remove_block(self, block) -> None:
        key = self.build_key(block)
        if not key:
            return

        mapping = self.get(self.cache_key)
        if mapping:
            mapping.pop(key, None)
        elif mapping is None:
            mapping = {}

        self.set(self.cache_key, mapping)

    def remove_pipeline(self, block, pipeline_uuid: str) -> None:
        key = self.build_key(block)
        if not key:
            return

        mapping = self.get(self.cache_key)
        if mapping is None:
            mapping = {}

        pipelines_dict = mapping.get(key, {})
        pipelines_dict.pop(pipeline_uuid, None)
        mapping[key] = pipelines_dict

        self.set(self.cache_key, mapping)

    async def initialize_cache_for_all_pipelines(self) -> None:
        from mage_ai.data_preparation.models.pipeline import Pipeline

        pipeline_uuids = Pipeline.get_all_pipelines(self.repo_path)

        pipeline_dicts = await asyncio.gather(
            *[Pipeline.load_metadata(uuid, raise_exception=False) for uuid in pipeline_uuids],
        )
        pipeline_dicts = [p for p in pipeline_dicts if p is not None]

        mapping = {}
        for pipeline_dict in pipeline_dicts:
            for block_dict in pipeline_dict.get('blocks', []):
                key = self.build_key(block_dict)
                if not key:
                    continue
                if key not in mapping:
                    mapping[key] = {}
                mapping[key][pipeline_dict['uuid']] = build_pipeline_dict(pipeline_dict)

        self.set(self.cache_key, mapping)



from dataclasses import dataclass, field
from enum import Enum
from typing import Dict

from mage_ai.shared.config import BaseConfig
from mage_ai.shared.logger import LoggingLevel


class LoggerType(str, Enum):
    DEFAULT = 'file'
    S3 = 's3'
    GCS = 'gcs'


@dataclass
class LoggingConfig(BaseConfig):
    type: LoggerType = LoggerType.DEFAULT
    level: LoggingLevel = LoggingLevel.INFO
    destination_config: Dict = field(default_factory=dict)
    # The period to keep the log files, e.g. '30d', '24h', '3w'
    retention_period: str = None



from mage_ai.services.redis.redis import init_redis_client
from mage_ai.settings import REDIS_URL


class DistributedLock:
    def __init__(
        self,
        lock_key_prefix='LOCK_KEY',
        lock_timeout=30,
    ):
        self.lock_key_prefix = lock_key_prefix
        self.lock_timeout = lock_timeout
        self.redis_client = init_redis_client(REDIS_URL)

    def __lock_key(self, key) -> str:
        return f'{self.lock_key_prefix}_{key}'

    def try_acquire_lock(self, key, timeout: int = None) -> bool:
        if not self.redis_client:
            return True
        # Try to acquire the lock by setting the lock key with a value and an expiration time
        acquired = self.redis_client.set(
            self.__lock_key(key),
            '1',
            nx=True,
            ex=timeout or self.lock_timeout,
        )
        return acquired is True

    def release_lock(self, key):
        # Release the lock by deleting the lock key
        if self.redis_client:
            self.redis_client.delete(self.__lock_key(key))



from mage_ai.shared.parsers import sample_output
from mage_ai.tests.base_test import TestCase


class ParsersTests(TestCase):
    def test_sample_output_empty_input(self):
        self.assertEqual(sample_output([]), ([], False))
        self.assertEqual(sample_output({}), ({}, False))

    def test_sample_output_no_sampling(self):
        input_data = [1, 2, 3]
        self.assertEqual(sample_output(input_data), (input_data, False))

        input_data = {'key': 'value', 'nested': {'inner': 'data'}}
        self.assertEqual(sample_output(input_data), (input_data, False))

    def test_sample_output_sampling_list(self):
        # A nested list with more than 20 items, should be sampled
        input_data = [i for i in range(30)]
        expected_output = ([i for i in range(20)], True)
        self.assertEqual(sample_output(input_data), expected_output)

    def test_sample_output_sampling_dict(self):
        input_data = {'key{}'.format(i): i for i in range(30)}
        expected_output = ({'key{}'.format(i): i for i in range(30)}, False)
        self.assertEqual(sample_output(input_data), expected_output)

    def test_sample_output_sampling_mixed(self):
        # A mixed structure with a nested list and dictionary, both exceeding the threshold
        input_data = {'key1': [i for i in range(30)], 'key2': {'inner_key': [i for i in range(30)]}}
        expected_output = (
            {
                'key1': [i for i in range(20)],
                'key2': {'inner_key': [i for i in range(20)]}
            },
            True,
        )
        self.assertEqual(sample_output(input_data), expected_output)

        input_data2 = {
            'key1': [
                {
                    'subkey_1': [i for i in range(30)],
                    'subkey_2': {
                        'subkey_21': {
                            'subkey_211': [i for i in range(30)]
                        }
                    },
                },
                {
                    'subkey_3': [1, 2, 3],
                }
            ]
        }
        expected_output2 = (
            {
                'key1': [
                    {
                        'subkey_1': [i for i in range(20)],
                        'subkey_2': {
                            'subkey_21': {
                                'subkey_211': [i for i in range(20)]
                            },
                        },
                    },
                    {
                        'subkey_3': [1, 2, 3],
                    }
                ]
            },
            True,
        )
        self.assertEqual(sample_output(input_data2), expected_output2)



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.cache.tag import TagCache


class TagResource(GenericResource):
    @classmethod
    async def collection(self, query, meta, user, **kwargs):
        cache = await TagCache.initialize_cache()
        tag_uuids = cache.get_tags().keys()

        return self.build_result_set(
            [dict(uuid=uuid) for uuid in tag_uuids],
            user,
            **kwargs,
        )
